unit MyRobot;

interface

uses
  SysUtils, Token;

type
  Point = Record
    x,y: Integer;
end;

type
  RPackageData = Record
    id: Integer;
    weight: Integer;
    verified: Boolean;
end;

type
  TMyRobot = Class (TObject)
  public
    constructor Create(Data: String);
    destructor Destroy(); override;
    procedure setLocationX(x: Integer);
    procedure setLocationY(y: Integer);
    procedure setReset();
    procedure setDest(pEnd: Point; path: String);
    procedure setDelivery(pDest: Point; path, list, wList: String);
    procedure setDeliveryHide(pDest: Point; path: String);
    procedure setStopGathering();
    procedure setOblivionOn(pDest: Point; path: String);
    procedure setOblivionOff();
    procedure dropDelivery();

    procedure onPathMove();
    function checkOnPath(): Boolean;

    procedure doMove(dir: Char);
    procedure doPick(id: Integer);
    procedure doDrop(id: Integer);

    function getMove(): String;
    function atDest(): Boolean;
    function waitOnPackage(): Boolean;
    function getStatus(): String;
    function packages(): String;
    function getWeight(): Integer;
    function getAvgDeliverCost(): Integer;
    function shouldGather(): Boolean;
    function pickupFailed(): Boolean;
  private
    dataReady: Boolean; // My Robot's Data is ready and valid!
    id: Integer;        // My Robot's Identifier
    cap: Integer;       // My Robot's Capacity
    money: Integer;     // My Robot's Money
    stMoney: Integer;   // My Robot's Starting money suply.
    prevMoney: Integer; // My Robot's Prev Money (money last time dropped packages)
    devCost: Integer;   // My Robot's cost to deliever last package set.
    dCostSum: Int64;    // Total cost to deliever all packages
    dCostCount: Integer;// Number of trips delieverd in;
    myLocation: Point;  // My Robot's Location
    myScore: Integer;   // My Robot's Score (Estimated);

    // State Info
    myState: Integer;                   // Current State
    myHaveDest: Boolean;                // Do I have a destination?
    myDest: Point;                      // Destination Point
    holdDest: Point;                    // Destination Holding Point
    myHaveGatherDest: Boolean;          // Do I have a gather destination?
    myGatherDest: Point;                // Gather Destination Point
    amGathering: Boolean;               // Am I gathering?
    myPath: String;                     // Path to destination
    myOnPath, onPathDis: Boolean;       // Am I on that Path?
    waitForPackage: Boolean;            // Am I waiting for the package info?
    packageList: String;                // My Package ID List
    myLastCommand: String;              // Last Command I sent to the server
    myLastAction: String;               // Last Action at reported by server
    myAlone: Boolean;                   // Am I alone? (no bot within 2 turns)
    myVeryAlone: Boolean;               // Am I very alone? (no bot within 12 turns)
    myCanBePushed: Boolean;             // Can I be pushed this round (robot within 1 turn)
    myRobotCount: Integer;              // Opponent Count
    myPushCount: Integer;               // Number of Times I've been pushed around

    // Oblivion Push
    oblivionPushOn: Boolean;            // Doing an Oblivion Push?
    myOblivionFinalPush: Integer;         // Price willing to pay for the final push
    myOblivionSetp: Integer;              // Price willing to pay per step

    // Package List
    myPackages: array of RPackageData;

    // Move Info
    myBid: Integer;     // Bid to Move
    myMove: String;     // Given Move
    myMoveData: String; // Move's Data
  published
    property identifier: Integer read id;
    property capacity: Integer read cap;
    property cash: Integer read money;
    property tripCost: Integer read devCost;
    property location: Point read myLocation;
    property score: Integer read myScore;
    property ready: Boolean read dataReady;
    property state: Integer read myState write myState;
    property haveDest: Boolean read myHaveDest write myHaveDest;
    property dest: Point read myDest;
    property haveGatherDest: Boolean read myHaveGatherDest write myHaveGatherDest;
    property gathering: Boolean read amGathering;
    property gatherDest: Point read myGatherDest;
    property onPath: Boolean read onPathDis;
    property path: String read myPath;
    property lastCommand: String read myLastCommand;
    property lastAction: String read myLastAction write myLastAction;
    property alone: Boolean read myAlone write myAlone;
    property veryAlone: Boolean read myVeryAlone write myVeryAlone;
    property pushAlone: Boolean read myCanBePushed write myCanBePushed;
    property oponents: Integer read myRobotCount write myRobotCount;
    property pushCount: Integer read myPushCount;
    property oblivionPush: Boolean read oblivionPushOn;
    property oblivionStep: Integer read myOblivionSetp;
    property oblivionKill: Integer read myOblivionFinalPush;
    property bid: Integer read myBid write myBid;
    property move: String read myMove write myMove;
    property moveData: String read myMoveData write myMoveData;
end;

const
  ROBOT_FINDPACKAGE = $00;
  ROBOT_DELIVER = $01;
  ROBOT_KILLER = $02;
  ROBOT_HIDE_GET = $03;
  ROBOT_HIDE_PUT = $04;
  ROBOT_OBLIVION = $02;
  STATE_STR: array[0..4] of String =
        ('Finding Package',
         'Dilivering Package',
         'Push into Oblivion',
         'Hide and Gather (Pick)',
         'Hide and Gather (Drop)');
  STMOVE_NORTH = 'S';
  STMOVE_EAST = 'E';
  STMOVE_SOUTH = 'N';
  STMOVE_WEST = 'W';
  MOVE_NORTH = $01;
  MOVE_EAST = $02;
  MOVE_SOUTH = $03;
  MOVE_WEST = $04;


implementation

constructor TMyRobot.Create(Data: String);
var
  myToken: TToken;
begin
  inherited Create();

  myAlone := false;
  myVeryAlone := false;
  myRobotCount := 0;

  myToken := TToken.Create(Data, ' ', '+');
  if (myToken.countTokens = 3) then begin
    id := StrToInt(myToken.nextToken());
    cap := StrToInt(myToken.nextToken());
    money := StrToInt(myToken.nextToken());
    prevMoney := money;
    stMoney := money;
    devCost := 0;
    dCostSum := 0;
    dCostCount := 0;
    dataReady := true;
    WriteLn('I''m playing as #' + IntToStr(id) + ', Capacity=' + IntToStr(cap) +
            ', and Money=' + IntToStr(money));
    myLocation.x := 0;
    myLocation.y := 0;
    myState := ROBOT_FINDPACKAGE;
    myHaveDest := false;
    myDest.x := 0;
    myDest.y := 0;
    holdDest.x := -1;
    holdDest.y := 0;
    myHaveGatherDest := false;
    myGatherDest.x := 0;
    myGatherDest.y := 0;
    amGathering := false;
    myPath := '';
    myOnPath := false;
    onPathDis := false;
    waitForPackage := false;
    myLastCommand := 'None';
    myLastAction := 'None';
  end
  else begin
    WriteLn('Malformed Player Data');
    dataReady := false;
  end;
end;

destructor TMyRobot.Destroy();
begin

  inherited Destroy();
end;

procedure TMyRobot.setLocationX(x: Integer);
begin
  myLocation.x := (x-1);
end;

procedure TMyRobot.setLocationY(y: Integer);
begin
  myLocation.y := (y-1);
end;

procedure TMyRobot.setReset();
begin
  myState := ROBOT_FINDPACKAGE;
  myHaveDest := false;
  myDest.x := 0;
  myDest.y := 0;
  holdDest.x := -1;
  holdDest.y := 0;
  myHaveGatherDest := false;
  myGatherDest.x := 0;
  myGatherDest.y := 0;
  amGathering := false;
  myPath := '';
  myOnPath := false;
  onPathDis := false;
  waitForPackage := false;
  myMove := 'Drop';
  myMoveData := '-1';
  myPushCount := 0;
end;

procedure TMyRobot.setDest(pEnd: Point; path: String);
begin
  if (length(path) > 0) and (path[1] <> '!') then begin
    myDest := pEnd;
    myPath := path;
    myHaveDest := true;
    myMove := 'Move';
    myMoveData := myPath[1];
    // WriteLn('Moving to ' + IntToStr(pEnd.x) + ',' + IntToStr(pEnd.y) + ' - ' + path);
    waitForPackage := false;
  end
  else begin
    myHaveDest := false;
    myMove := 'Drop';
    myMoveData := '-1';
    waitForPackage := false;
  end;
end;

procedure TMyRobot.setDelivery(pDest: Point; path, list, wList: String);
var
  count: Integer;
  idToken, weToken: TToken;
  i: Integer;
begin
  myDest := pDest;
  myPath := path;
  myHaveDest := false;
  myMove := 'Pick';
  myMoveData := list;
  // WriteLn('Delivering to ' + IntToStr(pDest.x) + ',' + IntToStr(pDest.y) + ' - ' + path);
  waitForPackage := false;
  packageList := list;
  myState := ROBOT_DELIVER;
  idToken := TToken.Create(list, ' ', '+');
  weToken := TToken.Create(wList, ' ', '+');
  count := idToken.countTokens();
  if (count > 0) and (count = weToken.countTokens) then begin
    setLength(myPackages, count);
    for i := low(myPackages) to high(myPackages) do begin
      myPackages[i].id := StrToInt(idToken.nextToken());
      myPackages[i].weight := StrToInt(weToken.nextToken());
      myPackages[i].verified := false;
    end;
  end;
  idToken.Free();
  weToken.Free();
end;

procedure TMyRobot.setDeliveryHide(pDest: Point; path: String);
begin
  if (holdDest.x = -1) then
    holdDest := myDest;
  myDest := pDest;
  myGatherDest := pDest;
  myPath := path;
  myHaveDest := false;
  waitForPackage := false;
  myHaveGatherDest := true;
  amGathering := true;
end;

procedure TMyRobot.setStopGathering();
begin
  myHaveGatherDest := false;
  amGathering := false;
  myDest := holdDest;
  holdDest.x := -1;
end;

procedure TMyRobot.setOblivionOn(pDest: Point; path: String);
var cost: Integer;
begin
  if (length(path) > 0) and (path[1] <> '!') then begin
    myDest := pDest;
    myPath := path;
    myHaveDest := true;
    myMove := 'Move';
    myMoveData := myPath[1];
    waitForPackage := false;
    // Set the price we're willing to pay to 150% the average cost
    // it takes to deliever packages.
    cost := (getAvgDeliverCost() * 3) div 2;
    if (not oblivionPushOn) then begin
      if (cost < 20) then begin
        myOblivionSetp := 5;
        myOblivionFinalPush := 9;
      end
      else begin
        // 66% of the total cost is spent on getting to water
        myOblivionSetp := ((cost * 2) div 3) div length(path);
        // The Rest is spent on the final death push
        myOblivionFinalPush := cost - (myOblivionSetp * length(path));
      end;
      oblivionPushOn := true;
      myState := ROBOT_OBLIVION;
    end;
  end
end;

procedure TMyRobot.setOblivionOff();
begin
  oblivionPushOn := false;
  myState := ROBOT_FINDPACKAGE;
  myMove := 'Drop';
  myMoveData := '-1';
end;

procedure TMyRobot.dropDelivery();
begin
  myHaveDest := false;
  amGathering := false;
  myMove := 'Drop';
  myMoveData := packageList;
  if (myState = ROBOT_HIDE_PUT) then
    myState := ROBOT_HIDE_GET
  else
    myState := ROBOT_FINDPACKAGE;
  packageList := '';
end;

procedure TMyRobot.onPathMove();
begin
  if (length(myPath) > 0) then begin
    myMove := 'Move';
    myMoveData := myPath[1];
    myPushCount := 0;
  end;
end;

function TMyRobot.checkOnPath(): Boolean;
begin
  Result := myOnPath;
  onPathDis := myOnPath;
  if (not myOnPath) then
    inc(myPushCount);
  myOnPath := false;
end;

procedure TMyRobot.doMove(dir: Char);
begin
  case (dir) of
    STMOVE_NORTH:       dec(myLocation.y);
    STMOVE_SOUTH:       inc(myLocation.y);
    STMOVE_WEST:        dec(myLocation.x);
    STMOVE_EAST:        inc(myLocation.x);
  end;
  if (length(myPath) > 0) and (dir = myPath[1]) then begin
    myPath := Copy(myPath, 2, length(myPath));
    myOnPath := true;
    onPathDis := true;
  end
  else begin
    myOnPath := false;
    onPathDis := false;
  end;
end;

procedure TMyRobot.doPick(id: Integer);
var i: Integer;
begin
  if (length(myPackages) > 0) then begin
    for i := low(myPackages) to high(myPackages) do begin
      if (myPackages[i].id = id) then
        myPackages[i].verified := true;
    end;
  end;
end;

procedure TMyRobot.doDrop(id: Integer);
var i: Integer;
begin
  if (length(myPackages) > 0) then begin
    for i := low(myPackages) to high(myPackages) do begin
      if (myPackages[i].id = id) then begin
        myPackages[i].id := -id;
        if (myState < ROBOT_HIDE_GET) and
           (myPackages[i].verified) and
           (location.x = dest.x) and
           (location.y = dest.y) then
          inc(myScore, myPackages[i].weight);
          if (prevMoney <> money) then begin
            devCost := prevMoney - money;
            inc(dCostSum, devCost);
            inc(dCostCount);
            prevMoney := money;
          end;
      end;
    end;
  end;
end;

function TMyRobot.getMove(): String;
begin
  dec(Money, abs(myBid));
  if (length(myMove) = 4) and (length(myMoveData) > 0) then
    Result := IntToStr(myBid) + ' ' + MyMove + ' ' + myMoveData
  else
    Result := '-1 Drop -1';
  myLastCommand := Result;
  { if (myMove = 'Move') then
    WriteLn('Me (' + IntToStr(myLocation.x) + ',' + IntToStr(myLocation.y) +
            ',' + IntToStr(Money) + ') - ' + STATE_STR[myState] + '::' +
            Result + ' (' + myPath + ')')
  else
    WriteLn('Me (' + IntToStr(myLocation.x) + ',' + IntToStr(myLocation.y) +
            ',' + IntToStr(Money) + ') - ' + STATE_STR[myState] + '::' + Result); }
end;

function TMyRobot.atDest(): Boolean;
begin
  Result := (myHaveDest) and (length(myPath) = 0);
end;

function TMyRobot.waitOnPackage(): Boolean;
begin
  Result := waitForPackage;
  waitForPackage := true;
end;

function TMyRobot.getStatus(): String;
begin
  Result := STATE_STR[myState];
end;

function TMyRobot.packages(): String;
var i, w: Integer;
begin
  if (length(myPackages) = 0) then begin
    Result := 'None';
  end
  else begin
    Result := '';
    w := 0;
    for i := low(myPackages) to high(myPackages) do begin
      if (myPackages[i].id > 0) and (myPackages[i].verified) then begin
        Result := IntToStr(myPackages[i].id) + '+' + ' ' + Result;
        inc(w, myPackages[i].weight);
      end
      else if (myPackages[i].id > 0) then begin
        Result := IntToStr(myPackages[i].id) + ' ' + Result;
      end;
    end;
    if (length(Result) > 0) then
      Result := Result + '(' + intToStr(w) + '/' + intToStr(capacity) + ')'
    else
      Result := 'None';
  end;
end;

function TMyRobot.getWeight(): Integer;
var i: Integer;
begin
  if (length(myPackages) = 0) then begin
    Result := 0;
  end
  else begin
    Result := 0;
    for i := low(myPackages) to high(myPackages) do begin
      if (myPackages[i].id > 0) and (myPackages[i].verified) then begin
        inc(Result, myPackages[i].weight);
      end
    end;
  end;
end;

function TMyRobot.getAvgDeliverCost(): Integer;
begin
  if (dCostCount = 0) then
    Result := 0
  else
    Result := dCostSum div dCostCount;
end;

// We should only hide and gather if we have more then 33% of our
// money left.
function TMyRobot.shouldGather(): Boolean;
begin
  Result := (money >= (stMoney div 3));
end;

// Determins if our picked command failed
function TMyRobot.pickupFailed(): Boolean;
var i: Integer;
begin
  Result := false;      // False by default
  if (length(myPackages) > 0) then begin
    for i := low(myPackages) to high(myPackages) do begin
      if (myPackages[i].id >= 0) and
         (not myPackages[i].verified) then
        Result := true;
    end;
  end;
end;

end.
