unit Network;

interface

uses
  SysUtils, Libc, IdGlobal, IdBaseComponent, IdComponent, IdTCPClient, Map;


type
  TRobotLoop = Class (TObject)
  public
    constructor Create();
  private
    myMap: TMap;
    tcpClient: TIdTCPClient;
    host: String;         // Host
    port: Integer;        // Port

    display: Boolean;

    procedure tcpOnConnect(Sender: TObject);
    procedure tcpOnDisconnect(Sender: TObject);
    procedure niceUI();
end;

implementation

const
  STATE_NOMAP = $0001;
  STATE_MAPDATA = $0002;
  STATE_PLAYERCFG = $0003;
  STATE_GAMESTART = $0004;
  STATE_GAMELOOP = $0005;
  DEBUG = false;

  DEFAULT_SERVER = 'localhost'; // }'icfp1.cse.ogi.edu';
  DEFAULT_PORT = 20005;  // } 22104;
  DEFAULT_SLEEP = 25;   // } 500;

constructor TRobotLoop.Create();
var
  dFile: TextFile;
begin
  if (ParamStr(1) = '') then
    host := DEFAULT_SERVER
  else
    host := ParamStr(1);
  if (ParamStr(2) = '') then
    port := DEFAULT_PORT
  else
    port := StrToInt(ParamStr(2));

  {$I-}
  AssignFile(dFile, GetCurrentDir() + '/NODISPLAY');
  FileMode := 0;
  Reset(dFile);
  CloseFile(dFile);
  {$I+}
  display := not (IOResult = 0);


  WriteLn('Connecting to ' + host + ':' + IntToStr(port));
  tcpClient := TIdTCPClient.Create(nil);
  tcpClient.Host := host;
  tcpClient.Port := port;
  tcpClient.OnConnected := tcpOnConnect;
  tcpClient.OnDisconnected := tcpOnDisconnect;
  tcpClient.MaxLineLength := 256*256 shr 2;
  tcpClient.Connect();

  // Check to see if we should display
end;

// We've connecting, send Player
procedure TRobotLoop.tcpOnConnect(Sender: TObject);
var
  myState: Integer;
  line: String;
  send: String;
  move: Boolean;
  linePack: String;
  lineGame: String;
  lineSys: String;
begin
  // Send Player
  tcpClient.WriteLn('Player');
  myState := STATE_NOMAP;
  move := true;

  // Read lines while we'll we're connected, be sure to close the socket
  // after we're done.

  while (tcpClient.Connected) do begin
    line := tcpClient.readLn();
    { if (DEBUG) then
      WriteLn('Read: ' + line); }
    case (myState) of
      STATE_NOMAP:      begin
                          myMap := TMap.Create(line);
                          myState := STATE_MAPDATA;
                        end;
      STATE_MAPDATA:    begin
                          myMap.ParseLine(line);
                          if (myMap.haveAllData) then
                            myState := STATE_PLAYERCFG;
                        end;
      STATE_PLAYERCFG:  begin
                          myMap.CreateMyPlayer(line);
                          myState := STATE_GAMESTART;
                        end;
      STATE_GAMESTART:  begin
                          myMap.ParseGameLine(line);
                          move := true;
                          myState := STATE_GAMELOOP;
                        end;
      STATE_GAMELOOP:   begin
                          // Check to see if this is player data or packages
                          move := (length(line) > 0) and (line[1] = '#');
                          if (length(line) > 0) and
                             ((line[1] = '*') or (line[1] = 'R')) then begin
                            writeLn('System: ' + line);
                            lineSys := line;
                          end
                          else if (move) then begin
                            myMap.ParseGameLine(line);
                            lineGame := line;
                          end
                          else begin
                            myMap.ParsePackageLine(line);
                            linePack := line;
                          end;
                        end;
    end;
    // Calculate and send our move
    if (myState = STATE_GAMELOOP) and (move) then begin
      myMap.robotAI();
      send := myMap.me.getMove();
      { if (DEBUG) then
        WriteLn('Send: ' + send); }
      tcpClient.WriteLn(send);
      if (display) and (myMap <> nil) and (myMap.Display) then begin
        niceUI();
        if (DEBUG) then begin
          WriteLn('Pack Line Length: ' + IntToStr(Length(linePack)));
          WriteLn('Game Line: ' + lineGame);
          WriteLn('System Line: ' + lineSys);
          WriteLn('PKDebug: ' + EOL + myMap.packDebug);
        end;
      end;
      // Sleep just a little
      if (display) then
        Sleep(DEFAULT_SLEEP);
    end;
  end;

  // Game's Over, close the connection and clean up!
  myMap.Free();
  myMap := nil;
  tcpClient.Free();
  tcpClient := nil;
end;

// Connection close, clean and exit
procedure TRobotLoop.tcpOnDisconnect(Sender: TObject);
begin
  WriteLn('Shutting down, final score: ' + IntToStr(myMap.me.score)
   + ', With ' + IntToStr(myMap.me.cash) + ' fule left');
end;

procedure TRobotLoop.niceUI();
begin
  Libc.system(pChar('clear'));
  WriteLn('{-- ' + host + ':' + IntToStr(port) + ' ----------- Robot #' +
        IntToStr(myMap.me.identifier) + ' --}');
  WriteLn(myMap.getDisplayLine1());
  WriteLn(myMap.getDisplayLine2());
  WriteLn(myMap.getDisplayLine3());
  WriteLn(myMap.getDisplayLine4());
  WriteLn(myMap.getDisplayLine5());
  WriteLn(myMap.getDisplayLine6());
  WriteLn(myMap.getDisplayLine7());
  WriteLn(myMap.getDisplayLine8());
  WriteLn(myMap.getDisplayLine9());
  myMap.DisplayMap(Min(myMap.mapWidth + 2, 79),Min(myMap.mapHeight, 12));
  { WriteLn('pkDebug:' + EOL + myMap.packDebug) }
  { if (myMap.me.state = $02) then begin
    WriteLn('Oblivion On: ' + myMap.boolToStr(myMap.me.oblivionPush));
    Sleep(5000);
  end; }
end;

end.
