unit Packages;

interface

uses
  SysUtils, IdGlobal, Token, myRobot;

type
  TPackages = class (TObject)
  public
    constructor Create(Data: String);
    destructor Destroy(); override;
    procedure buildPackageList(Data: String);

    procedure setDistance(loc: Point; dis: Integer);
    procedure picked(packageId: Integer);

    function getNextDest(): Point;
    function getBestPackageList(capacity: Integer; var pDest: Point;
        var pwList: String; var outWeight: Integer; home: Boolean): String;
    function getCount(): Integer;
    function getPickupWeight(capacity: Integer; home: Boolean): Integer;
  private
    id: array of Integer;               // Package ID's
    destX: array of Integer;            // Package Dest X
    destY: array of Integer;            // Package Dest Y
    weight: array of Integer;           // Package Weight
    dist: array of Integer;             // Package Distance from Location

    myDebug: String;
  published
    property debug: String read myDebug;
end;

implementation

constructor TPackages.Create(Data: String);
begin
  buildPackageList(data);
  myDebug := '';
end;

destructor TPackages.Destroy();
begin
  setLength(id, 0);
  setLength(destX, 0);
  setLength(destY, 0);
  setLength(weight, 0);
  setLength(dist, 0);
end;

procedure TPackages.buildPackageList(Data: String);
var
  i: Integer;
  myToken: TToken;
  count: Integer;
begin
  myToken := TToken.Create(Data, ' ', '+');
  count := myToken.countTokens();
  if (count > 0) and ((count mod 4) = 0) then begin
    setLength(id, count div 4);
    setLength(destX, count div 4);
    setLength(destY, count div 4);
    setLength(weight, count div 4);
    setLength(dist, count div 4);
    i := 0;
    while (myToken.hasMoreTokens()) do begin
      id[i] := StrToInt(myToken.nextToken());
      destX[i] := StrToInt(myToken.nextToken()) - 1;
      destY[i] := StrToInt(myToken.nextToken()) - 1;
      weight[i] := StrToInt(myToken.nextToken());
      dist[i] := -2;
      inc(i);
    end;
  end
  else if ((count mod 4) > 0) then begin
    WriteLn('Bad Package Data, Integer count incorrect');
  end;
end;

// Sets all packages to the given delivery point
procedure TPackages.setDistance(loc: Point; dis: Integer);
var i: Integer;
begin
  for i := 0 to high(id) do begin
    if (destX[i] = loc.x) and
       (destY[i] = loc.y) then begin
      dist[i] := dis;
    end;
  end;
end;

procedure TPackages.picked(packageId: Integer);
var i: Integer;
begin
  for i := 0 to high(id) do begin
    if (packageId = id[i]) then
      id[i] := -1;
  end;
end;

// Function returns our next dilivery point
function TPackages.getNextDest(): Point;
var i: Integer;
begin
  Result.x := -1;
  Result.y := -1;
  for i := 0 to high(id) do begin
    if (id[i] >= 0) and (dist[i] = -2) then begin
      Result.x := destX[i];
      Result.y := destY[i];
    end;
  end;
end;

function TPackages.getBestPackageList(capacity: Integer; var pDest: Point;
        var pwList: String; var outWeight: Integer; home: Boolean): String;
var
  i: Integer;
  lowDist: Integer;
  addWeight: Integer;
  lowWeight: Integer;
  lowWeightLoc: Integer;
  highWeight: Integer;
  highWeightLoc: Integer;
  added: Boolean;
  count: Integer;
  used: array of boolean;
begin
  lowDist := -1;
  Result := '';
  pwList := '';
  setLength(used, length(id));
  outWeight := 0;

  // find our shortest delivery route;
  for i := 0 to high(id) do begin
    if (lowDist = -1) or (dist[i] < lowDist) then begin
      if (dist[i] > 0) then begin
        lowDist := dist[i];
        pDest.x := destX[i];
        pDest.y := destY[i];
      end;
    end;
    myDebug := myDebug + EOL + 'Package ' + intToStr(i) + ' ID: ' +
    intToStr(id[i]) + ' -> (' +
    IntToStr(destX[i]) + ',' + IntToStr(destY[i]) + ',' +
    IntToStr(dist[i]) + ')' +
    ', Weight: ' + IntToStr(weight[i]);
    used[i] := false;
  end;
  // myDebug := 'Shortest Delivery Distance: ' + IntToStr(lowDist);

  // Now find the lowest weight packages we can pick up. In this game
  // smaller is better, if you get pushed the package you drop won't
  // be of as much effect if you get the larger packages
  // However, if we only find one before we're over capacity
  // take the heavest.
  addWeight := 0;
  count := 0;
  added := true;
  while (added) do begin
    added := false;
    lowWeight := -1;
    lowWeightLoc := -1;
    for i := 0 to high(id) do begin
      if (dist[i] = lowDist) and (id[i] >= 0) and (not used[i]) and
         (destX[i] = pDest.x) and (destY[i] = pDest.y) and
         ((lowWeight = -1) or (weight[i] < lowWeight)) and
         (weight[i]+addWeight <= capacity) then begin
        lowWeight := weight[i];
        lowWeightLoc := i;
        added := true;
      end;
    end;
    if (added) then begin
      inc(count);
      inc(addWeight, weight[lowWeightLoc]);
      outWeight := addWeight;
      if (Result = '') then begin
        Result := IntToStr(id[lowWeightLoc]);
        pwList := IntToStr(weight[lowWeightLoc]);
      end
      else begin
        Result := Result + ' ' + IntToStr(id[lowWeightLoc]);
        pwList := pwList + ' ' + IntToStr(weight[lowWeightLoc]);
      end;
      used[lowWeightLoc] := true;
    end;
  end;
  // Only found one, take the heavest package
  if (count = 1) then begin
    highWeight := -1;
    highWeightLoc := -1;
    outWeight := 0;
    for i := 0 to high(id) do begin
      if (dist[i] = lowDist) and (id[i] >= 0) and
         ((highWeight = -1) or (weight[i] > highWeight)) and
         (weight[i] < capacity) then begin
        highWeight := weight[i];
        highWeightLoc := i;
      end
      // fix for some odd behavior, if weight = capacity it can
      // only be picked up from a home base. ?????
      else if (not home) and (weight[i] = capacity) then
        id[i] := -1;
    end;
    if (highWeightLoc = -1) then begin
      Result := '';
      pwList := '';
    end
    else begin
      Result := IntToStr(id[highWeightLoc]);
      pwList := IntToStr(weight[highWeightLoc]);
      outWeight := weight[highWeightLoc];
    end;
  end;
end;

function TPackages.getCount(): Integer;
var i: Integer;
begin
  Result := 0;
  for i := 0 to high(id) do begin
    if (id[i] >= 0) then
      inc(Result);
  end;
end;

function TPackages.getPickupWeight(capacity: Integer; home: Boolean): Integer;
var
  pt: Point;
  pe: String;
begin
  self.getBestPackageList(capacity, pt, pe, Result, home);
end;

end.
