unit Token;

interface

  uses SysUtils;

  type
    TToken = Class (TObject)
  private
    { private members }
    Base, Current, Delim, Escape: String;
    Loc: Integer;
  public
    { public members }
    constructor Create(S, T, E: String);
    destructor Destroy(); override;
    function nextToken(): String;
    function hasMoreTokens(): Boolean;
    function countTokens(): Integer;
    procedure Reset();
  end;

implementation

constructor TToken.Create(S, T, E: String);
begin
  inherited Create();
  Base := S;
  Current := S;
  Delim := T;
  Escape := E;
end;

destructor TToken.Destroy();
begin
  setLength(Base, 0);
  setLength(Current, 0);
  setLength(Delim, 0);
  setLength(Escape, 0);

  inherited Destroy();
end;

function TToken.nextToken(): String;
var
  S: String;
begin
  if not (Pos(Delim, Current) = 0) then begin
    S := Copy(Current, Loc, Pos(Delim, Current)-1);
    Current := Copy(Current, Pos(Delim, Current)+Length(Delim), Length(Current));
    Result := S;
  end else begin
    S := Current;
    Current := '';
    Result := S;
  end;
  Result := StringReplace(Result, Escape, Delim, [rfReplaceAll]);
end;

function TToken.hasMoreTokens(): Boolean;
begin
  if (Length(Current) = 0) then
    Result := False
  else
    Result := True
end;

function TToken.countTokens(): Integer;
var
  data: String;
begin
  Result := 1;
  data := nextToken();
  while hasMoreTokens() do begin
    inc(Result);
    data := nextToken();
  end;
  Reset();
end;

procedure TToken.Reset();
begin
  Current := Base;
  Loc := 0;
end;

end.
